@@ -1,3 +1,4 @@
+use std::error::Error;
 use std::{cmp, fs, io};
 use std::collections::{HashMap, HashSet};
 use std::path::{Path, PathBuf};
@@ -11,6 +12,7 @@ use rpki::rrdp::{DeltaInfo, DeltaListError, NotificationFile};
 use tempfile::NamedTempFile;
 use crate::config::Config;
 use crate::error::{Fatal, RunFailed};
+use crate::log::LogBookWriter;
 use crate::metrics::{Metrics, RrdpRepositoryMetrics};
 use crate::utils::fatal;
 use crate::utils::archive::{ArchiveError, OpenError};
@@ -18,7 +20,9 @@ use crate::utils::dump::DumpRegistry;
 use crate::utils::json::JsonBuilder;
 use crate::utils::sync::{Mutex, RwLock};
 use crate::utils::uri::UriExt;
-use super::archive::{FallbackTime, RrdpArchive, RepositoryState};
+use super::archive::{
+    FallbackTime, RrdpArchive, RepositoryState, SnapshotRrdpArchive,
+};
 use super::http::{HttpClient, HttpStatus};
 use super::update::{
     DeltaUpdate, Notification, SnapshotError, SnapshotReason, SnapshotUpdate
@@ -117,7 +121,7 @@ impl Collector {
                         );
                         return Err(Fatal)
                     }
-                    Err(OpenError::Archive(ArchiveError::Corrupt)) => {
+                    Err(OpenError::Archive(ArchiveError::Corrupt(_))) => {
                         match fs::remove_file(entry.path()) {
                             Ok(()) => {
                                 info!(
@@ -142,14 +146,17 @@ impl Collector {
         Ok(())
     }
 
-    pub fn start(&self) -> Run {
+    pub fn start(&self) -> Run<'_> {
         Run::new(self)
     }
 
     #[allow(clippy::mutable_key_type)]
     pub fn dump(&self, dir: &Path) -> Result<(), Fatal> {
         let dir = dir.join("rrdp");
-        debug!("Dumping RRDP collector content to {}", dir.display());
+        debug!("Dumping RRDP collector content from {} to {}", 
+            self.working_dir.display(), 
+            dir.display()
+        );
         let mut registry = DumpRegistry::new(dir);
         let mut states = HashMap::new();
         for entry in fatal::read_dir(&self.working_dir)? {
@@ -360,21 +367,20 @@ impl<'a> Run<'a> {
     /// This just downloads the file. It is not cached since that is done
     /// by the store anyway.
     pub fn load_ta(&self, uri: &uri::Https) -> Option<Bytes> {
-        let mut response = match self.collector.http.response(uri, false) {
+        let mut response = match self.collector.http.response(uri) {
             Ok(response) => response,
             Err(_) => return None,
         };
         if response.content_length() > self.collector.config().max_object_size {
             warn!(
-                "Trust anchor certificate {} exceeds size limit. \
-                 Ignoring.",
-                uri
+                "Trust anchor certificate {uri} exceeds size limit. \
+                 Ignoring."
             );
             return None
         }
         let mut bytes = Vec::new();
         if let Err(err) = response.copy_to(&mut bytes) {
-            info!("Failed to get trust anchor {}: {}", uri, err);
+            info!("Failed to get trust anchor {uri}: {err}");
             return None
         }
         Some(Bytes::from(bytes))
@@ -397,10 +403,10 @@ impl<'a> Run<'a> {
     /// is the first attempt at updating the repository.
     pub fn load_repository(
         &self, rpki_notify: &uri::Https
-    ) -> Result<(LoadResult, bool), RunFailed> {
+    ) -> Result<LoadResult, RunFailed> {
         // If we already tried updating, we can return already.
         if let Some(repo) = self.updated.read().get(rpki_notify) {
-            return Ok((repo.read()?, false))
+            return repo.read()
         }
 
         // Get a clone of the (arc-ed) mutex. Make a new one if there isn’t
@@ -416,29 +422,39 @@ impl<'a> Run<'a> {
         let _lock = mutex.lock();
         if let Some(repo) = self.updated.read().get(rpki_notify) {
             self.running.write().remove(rpki_notify);
-            return Ok((repo.read()?, false))
+            return repo.read()
         }
 
+        let mut log = LogBookWriter::new(
+            self.collector.config.log_repository_issues.then(|| {
+                format!("RRDP {}: ", rpki_notify)
+            })
+        );
+
         // Now we can update the repository. But we only do this if we like
         // the URI.
-        let (repo, metrics) = if
+        let (repo, mut metrics) = if
             self.collector.config().filter_dubious
             && rpki_notify.has_dubious_authority()
         {
             let mut metrics = RrdpRepositoryMetrics::new(rpki_notify.clone());
             metrics.notify_status = HttpStatus::Rejected;
-            warn!(
-                "{}: Dubious host name. Not using the repository.",
-                rpki_notify
-            );
+            log.warn(format_args!(
+                "Dubious host name. Not using the repository."
+            ));
             (LoadResult::Unavailable, metrics)
         }
         else {
             RepositoryUpdate::new(
-                self.collector, rpki_notify
+                self.collector, rpki_notify, &mut log,
             )?.try_update()?
         };
 
+        let log = log.into_book();
+        if !log.is_empty() {
+            metrics.log_book = Some(log);
+        }
+
         // Insert metrics.
         self.metrics.lock().push(metrics);
 
@@ -450,7 +466,7 @@ impl<'a> Run<'a> {
         // Remove from running.
         self.running.write().remove(rpki_notify);
 
-        Ok((res, true))
+        Ok(res)
     }
 
     #[allow(clippy::mutable_key_type)]
@@ -603,6 +619,9 @@ pub struct RrdpConfig {
 
     /// The maximum length of the delta list in a notification file.
     pub max_delta_list_len: usize,
+
+    /// Log issues also to the process log?
+    pub log_repository_issues: bool,
 }
 
 impl<'a> From<&'a Config> for RrdpConfig {
@@ -613,6 +632,7 @@ impl<'a> From<&'a Config> for RrdpConfig {
             max_object_size: config.max_object_size,
             max_delta_count: config.rrdp_max_delta_count,
             max_delta_list_len: config.rrdp_max_delta_list_len,
+            log_repository_issues: config.log_repository_issues,
         }
     }
 }
@@ -719,17 +739,21 @@ struct RepositoryUpdate<'a> {
     path: Arc<PathBuf>,
     rpki_notify: &'a uri::Https,
     metrics: RrdpRepositoryMetrics,
+    log: &'a mut LogBookWriter,
 }
 
 impl<'a> RepositoryUpdate<'a> {
     fn new(
-        collector: &'a Collector, rpki_notify: &'a uri::Https,
+        collector: &'a Collector,
+        rpki_notify: &'a uri::Https,
+        log: &'a mut LogBookWriter,
     ) -> Result<Self, RunFailed> {
         Ok(Self {
             collector,
             path: Arc::new(collector.repository_path(rpki_notify)?),
             rpki_notify,
             metrics: RrdpRepositoryMetrics::new(rpki_notify.clone()),
+            log,
         })
     }
 
@@ -775,18 +799,15 @@ impl<'a> RepositoryUpdate<'a> {
             LoadResult::Current
         }
         else if let Some(date) = best_before {
-            info!(
-                "RRDP {}: Update failed and \
-                current copy is expired since {}.",
-                self.rpki_notify, date
-            );
+            self.log.info(format_args!(
+                "Update failed and current copy is expired since {date}.",
+            ));
             LoadResult::Stale
         }
         else {
-            info!(
-                "RRDP {}: Update failed and there is no current copy.",
-                self.rpki_notify
-            );
+            self.log.info(format_args!(
+                "Update failed and there is no current copy."
+            ));
             LoadResult::Unavailable
         };
         Ok((res, self.metrics))
@@ -804,6 +825,7 @@ impl<'a> RepositoryUpdate<'a> {
             current.as_ref().map(|x| &x.1),
             &mut self.metrics.notify_status,
             self.collector.config.max_delta_list_len,
+            self.log,
         ) {
             Ok(Some(notify)) => notify,
             Ok(None) => {
@@ -837,7 +859,7 @@ impl<'a> RepositoryUpdate<'a> {
         &mut self,
         current: Option<(RrdpArchive, RepositoryState)>,
     ) -> Result<(), RunFailed> {
-        info!("RRDP {}: Not modified.", self.rpki_notify);
+        self.log.info(format_args!("Not modified."));
         if let Some((mut archive, mut state)) = current {
             // Copy serial and session to the metrics so they will still be
             // present.
@@ -857,21 +879,54 @@ impl<'a> RepositoryUpdate<'a> {
         &mut self,
         notify: &Notification,
     ) -> Result<bool, RunFailed> {
-        debug!("RRDP {}: updating from snapshot.", self.rpki_notify);
+        self.log.debug(format_args!("updating from snapshot."));
         let (file, path) = self.collector.temp_file()?;
-        let mut archive = RrdpArchive::create_with_file(file, path.clone())?;
+        let mut archive = SnapshotRrdpArchive::create_with_file(
+            file, path.clone()
+        )?;
         if let Err(err) = SnapshotUpdate::new(
-            self.collector, &mut archive, notify, &mut self.metrics
+            self.collector, &mut archive, notify, &mut self.metrics,
         ).try_update() {
+            // XXX This should probably be done nicer. The problem is that I
+            // cannot peek into the response stream to see if it has a chance
+            // of being processable (e.g. is empty). This means that it will
+            // end up as an XML error, which whilst technically true is not
+            // really helpful in the case of a broken/timed out HTTP stream.
             if let SnapshotError::RunFailed(err) = err {
-                debug!("RRDP {}: snapshot update failed.", self.rpki_notify);
+                self.log.debug(format_args!("snapshot update failed."));
                 return Err(err)
             }
+            else if let SnapshotError::Http(err) = err {
+                if let Some(source) = err.source() {
+                    warn!(
+                        "RRDP {}: Failed to process snapshot file {}: {} ({})",
+                        self.rpki_notify, notify.content().snapshot().uri(), 
+                        err, source
+                    );
+                } else {
+                    warn!(
+                        "RRDP {}: Failed to process snapshot file {}: {}",
+                        self.rpki_notify, notify.content().snapshot().uri(), 
+                        err
+                    );
+                }
+                return Ok(false)
+            } 
+            else if let SnapshotError::Rrdp(err) = err {
+
+                let mut err: &dyn Error = &err;
+                while let Some(e) = err.source() {
+                    err = e;
+                }
+                warn!("RRDP {}: Failed to process snapshot file XML {}: {}", 
+                    self.rpki_notify, notify.content().snapshot().uri(), err);
+                return Ok(false);
+            }
             else {
-                warn!(
-                    "RRDP {}: failed to process snapshot file {}: {}",
-                    self.rpki_notify, notify.content().snapshot().uri(), err
-                );
+                self.log.warn(format_args!(
+                    "failed to process snapshot file {}: {}",
+                    notify.content().snapshot().uri(), err
+                ));
                 return Ok(false)
             }
         }
@@ -899,7 +954,7 @@ impl<'a> RepositoryUpdate<'a> {
             return Err(RunFailed::fatal())
         }
 
-        debug!("RRDP {}: snapshot update completed.", self.rpki_notify);
+        self.log.debug(format_args!("snapshot update completed."));
         Ok(true)
     }
 
@@ -918,10 +973,9 @@ impl<'a> RepositoryUpdate<'a> {
         if let Err(err) = notify.content().delta_status() {
             match err {
                 DeltaListError::Oversized => {
-                    info!(
-                        "RRDP {}: Overly large delta set in notification file",
-                        self.rpki_notify
-                    );
+                    self.log.info(format_args!(
+                        "Overly large delta set in notification file",
+                    ));
                     return Ok(Some(SnapshotReason::LargeDeltaSet));
                 }
             }
@@ -939,19 +993,17 @@ impl<'a> RepositoryUpdate<'a> {
         if !deltas.is_empty() {
             let count = deltas.len();
             for (i, info) in deltas.iter().enumerate() {
-                debug!(
-                    "RRDP {}: Delta update step ({}/{}).",
-                    self.rpki_notify, i + 1, count
-                );
+                self.log.debug(format_args!(
+                    "Delta update step ({}/{}).", i + 1, count
+                ));
                 if let Err(err) = DeltaUpdate::new(
                     self.collector, &mut archive,
                     notify.content().session_id(),
                     info, &mut self.metrics
                 ).try_update() {
-                    warn!(
-                        "RRDP {}: failed to process delta: {}",
-                        self.rpki_notify, err,
-                    );
+                    self.log.warn(format_args!(
+                        "failed to process delta: {}", err,
+                    ));
                     return Ok(Some(SnapshotReason::ConflictingDelta))
                 }
             }
@@ -971,7 +1023,7 @@ impl<'a> RepositoryUpdate<'a> {
             }
         }
 
-        debug!("RRDP {}: Delta update completed.", self.rpki_notify);
+        self.log.debug(format_args!("Delta update completed."));
         Ok(None)
     }
 
@@ -980,17 +1032,19 @@ impl<'a> RepositoryUpdate<'a> {
     /// Returns an empty slice if no update is necessary.
     /// Returns a non-empty slice of the sequence of deltas to be applied.
     fn calc_deltas<'b>(
-        &self,
+        &mut self,
         notify: &'b NotificationFile,
         state: &RepositoryState
     ) -> Result<&'b [DeltaInfo], SnapshotReason> {
         if notify.session_id() != state.session {
-            debug!("New session. Need to get snapshot.");
+            self.log.debug(format_args!(
+                "New session. Need to get snapshot."
+            ));
             return Err(SnapshotReason::NewSession)
         }
-        debug!("{}: Serials: us {}, them {}.",
-            self.rpki_notify, state.serial, notify.serial()
-        );
+        self.log.debug(format_args!(
+            "Serials: us {}, them {}.", state.serial, notify.serial()
+        ));
         if notify.serial() == state.serial {
             return Ok(&[]);
         }
@@ -1002,7 +1056,9 @@ impl<'a> RepositoryUpdate<'a> {
         if notify.deltas().last().map(|delta| delta.serial())
             != Some(notify.serial())
         {
-            debug!("Last delta serial differs from current serial.");
+            self.log.debug(format_args!(
+                "Last delta serial differs from current serial."
+            ));
             return Err(SnapshotReason::BadDeltaSet)
         }
 
@@ -1015,13 +1071,15 @@ impl<'a> RepositoryUpdate<'a> {
             let first = match deltas.first() {
                 Some(first) => first,
                 None => {
-                    debug!("Ran out of deltas.");
+                    self.log.debug(format_args!("Ran out of deltas."));
                     return Err(SnapshotReason::BadDeltaSet)
                 }
             };
             match first.serial().cmp(&serial) {
                 cmp::Ordering::Greater => {
-                    debug!("First delta is too new ({})", first.serial());
+                    self.log.debug(format_args!(
+                        "First delta is too new ({})", first.serial()
+                    ));
                     return Err(SnapshotReason::OutdatedLocal)
                 }
                 cmp::Ordering::Equal => break,
@@ -1030,10 +1088,9 @@ impl<'a> RepositoryUpdate<'a> {
         }
 
         if deltas.len() > self.collector.config.max_delta_count {
-            debug!(
-                "RRDP: {}: Too many delta steps required ({})",
-                self.rpki_notify, deltas.len()
-            );
+            self.log.debug(format_args!(
+                "Too many delta steps required ({})", deltas.len()
+            ));
             return Err(SnapshotReason::TooManyDeltas)
         }
 
