@@ -1,4 +1,5 @@
 
+use std::error::Error;
 use std::{error, fmt, io};
 use std::collections::HashSet;
 use std::io::Read;
@@ -7,14 +8,17 @@ use chrono::{DateTime, Utc};
 use log::{error, warn};
 use reqwest::StatusCode;
 use ring::digest;
-use ring::constant_time::verify_slices_are_equal;
 use rpki::{rrdp, uri};
 use rpki::rrdp::{DeltaInfo, NotificationFile, ProcessDelta, ProcessSnapshot};
 use uuid::Uuid;
 use crate::error::{Failed, RunFailed};
+use crate::log::LogBookWriter;
 use crate::metrics::RrdpRepositoryMetrics;
 use crate::utils::archive::{ArchiveError, PublishError};
-use super::archive::{AccessError, FallbackTime, RepositoryState, RrdpArchive};
+use super::archive::{
+    AccessError, FallbackTime, RepositoryState, RrdpArchive,
+    SnapshotRrdpArchive,
+};
 use super::base::Collector;
 use super::http::{HttpClient, HttpResponse, HttpStatus};
 
@@ -51,19 +55,23 @@ impl Notification {
         state: Option<&RepositoryState>,
         status: &mut HttpStatus,
         delta_list_limit: usize,
+        log: &mut LogBookWriter,
     ) -> Result<Option<Self>, Failed> {
         let response = match http.conditional_response(
             uri,
             state.and_then(|state| state.etag.as_ref()),
             state.and_then(|state| state.last_modified()),
-            true
         ) {
             Ok(response) => {
                 *status = response.status().into();
                 response
             }
             Err(err) => {
-                warn!("RRDP {}: {}", uri, err);
+                if let Some(source) = err.source() {
+                    log.warn(format_args!("{err} ({source})"));
+                } else {
+                    log.warn(format_args!("{err}"));
+                }
                 *status = HttpStatus::Error;
                 return Err(Failed)
             }
@@ -73,15 +81,15 @@ impl Notification {
             Ok(None)
         }
         else if response.status() != StatusCode::OK {
-            warn!(
-                "RRDP {}: Getting notification file failed with status {}",
-                uri, response.status()
-            );
+            log.warn(format_args!(
+                "Getting notification file failed with status {}",
+                response.status()
+            ));
             Err(Failed)
         }
         else {
             Notification::from_response(
-                uri.clone(), response, delta_list_limit
+                uri.clone(), response, delta_list_limit, log,
             ).map(Some)
         }
     }
@@ -91,20 +99,23 @@ impl Notification {
     ///
     /// Assumes that the response status was 200 OK.
     fn from_response(
-        uri: uri::Https, response: HttpResponse, delta_list_limit: usize
+        uri: uri::Https,
+        response: HttpResponse,
+        delta_list_limit: usize,
+        log: &mut LogBookWriter,
     ) -> Result<Self, Failed> {
         let etag = response.etag();
         let last_modified = response.last_modified();
         let mut content = NotificationFile::parse_limited(
             io::BufReader::new(response), delta_list_limit
         ).map_err(|err| {
-            warn!("RRDP {}: {}", uri, err);
+            log.warn(format_args!("{err}"));
             Failed
         })?;
         if !content.has_matching_origins(&uri) {
-            warn!("RRDP {}: snapshot or delta files with different origin",
-                uri
-            );
+            log.warn(format_args!(
+                "snapshot or delta files with different origin"
+            ));
             return Err(Failed)
         }
         content.sort_deltas();
@@ -165,7 +176,7 @@ pub struct SnapshotUpdate<'a> {
     collector: &'a Collector,
 
     /// The archive to store the snapshot into.
-    archive: &'a mut RrdpArchive,
+    archive: &'a mut SnapshotRrdpArchive,
 
     /// The notification file pointing to the snapshot.
     notify: &'a Notification,
@@ -177,7 +188,7 @@ pub struct SnapshotUpdate<'a> {
 impl<'a> SnapshotUpdate<'a> {
     pub fn new(
         collector: &'a Collector,
-        archive: &'a mut RrdpArchive,
+        archive: &'a mut SnapshotRrdpArchive,
         notify: &'a Notification,
         metrics: &'a mut RrdpRepositoryMetrics,
     ) -> Self {
@@ -186,7 +197,7 @@ impl<'a> SnapshotUpdate<'a> {
 
     pub fn try_update(mut self) -> Result<(), SnapshotError> {
         let response = match self.collector.http().response(
-            self.notify.content.snapshot().uri(), false
+            self.notify.content.snapshot().uri()
         ) {
             Ok(response) => {
                 self.metrics.payload_status = Some(response.status().into());
@@ -205,18 +216,15 @@ impl<'a> SnapshotUpdate<'a> {
 
         let mut reader = io::BufReader::new(HashRead::new(response));
         self.process(&mut reader)?;
-        let hash = reader.into_inner().into_hash();
-        if verify_slices_are_equal(
-            hash.as_ref(),
-            self.notify.content.snapshot().hash().as_ref()
-        ).is_err() {
-            return Err(SnapshotError::HashMismatch)
-        }
+        reader.into_inner().verify_hash(
+            self.notify.content.snapshot().hash()
+        )?;
         self.archive.publish_state(
             &self.notify.to_repository_state(
                 self.collector.config().fallback_time
             )
         )?;
+        self.archive.finalize()?;
         Ok(())
     }
 }
@@ -256,7 +264,7 @@ impl ProcessSnapshot for SnapshotUpdate<'_> {
             PublishError::AlreadyExists => {
                 SnapshotError::DuplicateObject(uri.clone())
             }
-            PublishError::Archive(ArchiveError::Corrupt) => {
+            PublishError::Archive(ArchiveError::Corrupt(_)) => {
                 warn!(
                     "Temporary RRDP repository file {} became corrupt.",
                     self.archive.path().display(),
@@ -323,7 +331,7 @@ impl<'a> DeltaUpdate<'a> {
 
     pub fn try_update(mut self) -> Result<(), DeltaError> {
         let response = match self.collector.http().response(
-            self.info.uri(), false
+            self.info.uri()
         ) {
             Ok(response) => {
                 self.metrics.payload_status = Some(response.status().into());
@@ -342,13 +350,7 @@ impl<'a> DeltaUpdate<'a> {
 
         let mut reader = io::BufReader::new(HashRead::new(response));
         self.process(&mut reader)?;
-        let hash = reader.into_inner().into_hash();
-        if verify_slices_are_equal(
-            hash.as_ref(),
-            self.info.hash().as_ref()
-        ).is_err() {
-            return Err(DeltaError::DeltaHashMismatch)
-        }
+        reader.into_inner().verify_hash(self.info.hash())?;
         Ok(())
     }
 }
@@ -458,11 +460,16 @@ impl<R> HashRead<R> {
         }
     }
 
-    /// Converts the reader into the hash.
-    pub fn into_hash(self) -> rrdp::Hash {
-        // Unwrap should be safe: This can only fail if the slice has the
-        // wrong length.
-        rrdp::Hash::try_from(self.context.finish()).unwrap()
+    /// Checks that the hash matches the provided hash.
+    pub fn verify_hash(
+        self, expected: rrdp::Hash
+    ) -> Result<(), HashMismatch> {
+        if self.context.finish().as_ref() != expected.as_ref() {
+            Err(HashMismatch)
+        }
+        else {
+            Ok(())
+        }
     }
 }
 
@@ -545,9 +552,7 @@ impl<R: io::Read> io::Read for RrdpDataRead<'_, R> {
             Ok(res) => res,
             Err(err) => {
                 self.err = Some(RrdpDataReadError::Read(err));
-                return Err(io::Error::new(
-                    io::ErrorKind::Other, "reading data failed",
-                ))
+                return Err(io::Error::other("reading data failed"))
             }
         };
         if let Some(left) = self.left {
@@ -560,9 +565,7 @@ impl<R: io::Read> io::Read for RrdpDataRead<'_, R> {
                     self.err = Some(
                         RrdpDataReadError::LargeObject(self.uri.clone())
                     );
-                    return Err(io::Error::new(
-                        io::ErrorKind::Other, "size limit exceeded"
-                    ))
+                    return Err(io::Error::other("size limit exceeded"))
                 }
             };
             if res64 > left {
@@ -570,9 +573,7 @@ impl<R: io::Read> io::Read for RrdpDataRead<'_, R> {
                 self.err = Some(
                     RrdpDataReadError::LargeObject(self.uri.clone())
                 );
-                Err(io::Error::new(
-                    io::ErrorKind::Other, "size limit exceeded")
-                )
+                Err(io::Error::other("size limit exceeded"))
             }
             else {
                 self.left = Some(left - res64);
@@ -659,6 +660,12 @@ enum RrdpDataReadError {
 }
 
 
+//------------ HashMismatch --------------------------------------------------
+
+/// The hash of a snapshot or delta didn’t match the expected value.
+struct HashMismatch;
+
+
 //------------ SnapshotError -------------------------------------------------
 
 /// An error happened during snapshot processing.
@@ -728,38 +735,42 @@ impl From<RrdpDataReadError> for SnapshotError {
     }
 }
 
+impl From<HashMismatch> for SnapshotError {
+    fn from(_: HashMismatch) -> Self {
+        Self::HashMismatch
+    }
+}
+
 impl fmt::Display for SnapshotError {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match *self {
             SnapshotError::Http(ref err) => err.fmt(f),
             SnapshotError::HttpStatus(status) => {
-                write!(f, "HTTP {}", status)
+                write!(f, "HTTP {status}")
             }
             SnapshotError::Rrdp(ref err) => err.fmt(f),
             SnapshotError::SessionMismatch { ref expected, ref received } => {
                 write!(
                     f,
-                    "session ID mismatch (notification_file: {}, \
-                     snapshot file: {}",
-                     expected, received
+                    "session ID mismatch (notification_file: {expected}, \
+                     snapshot file: {received}"
                 )
             }
             SnapshotError::SerialMismatch { ref expected, ref received } => {
                 write!(
                     f,
-                    "serial number mismatch (notification_file: {}, \
-                     snapshot file: {}",
-                     expected, received
+                    "serial number mismatch (notification_file: {expected}, \
+                     snapshot file: {received}"
                 )
             }
             SnapshotError::DuplicateObject(ref uri) => {
-                write!(f, "duplicate object: {}", uri)
+                write!(f, "duplicate object: {uri}")
             }
             SnapshotError::HashMismatch => {
                 write!(f, "hash value mismatch")
             }
             SnapshotError::LargeObject(ref uri) => {
-                write!(f, "object exceeds size limit: {}", uri)
+                write!(f, "object exceeds size limit: {uri}")
             }
             SnapshotError::RunFailed(_) => Ok(()),
         }
@@ -843,62 +854,63 @@ impl From<RrdpDataReadError> for DeltaError {
     }
 }
 
+impl From<HashMismatch> for DeltaError {
+    fn from(_: HashMismatch) -> Self {
+        Self::DeltaHashMismatch
+    }
+}
+
 impl fmt::Display for DeltaError {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match *self {
             DeltaError::Http(ref err) => err.fmt(f),
             DeltaError::HttpStatus(status) => {
-                write!(f, "HTTP {}", status)
+                write!(f, "HTTP {status}")
             }
             DeltaError::Rrdp(ref err) => err.fmt(f),
             DeltaError::SessionMismatch { ref expected, ref received } => {
                 write!(
                     f,
-                    "session ID mismatch (notification_file: {}, \
-                     snapshot file: {}",
-                     expected, received
+                    "session ID mismatch (notification_file: {expected}, \
+                     snapshot file: {received}"
                 )
             }
             DeltaError::SerialMismatch { ref expected, ref received } => {
                 write!(
                     f,
-                    "serial number mismatch (notification_file: {}, \
-                     snapshot file: {}",
-                     expected, received
+                    "serial number mismatch (notification_file: {expected}, \
+                     snapshot file: {received}"
                 )
             }
             DeltaError::MissingObject { ref uri } => {
                 write!(
                     f,
-                    "reference to missing object {}",
-                    uri
+                    "reference to missing object {uri}"
                 )
             }
             DeltaError::ObjectAlreadyPresent { ref uri } => {
                 write!(
                     f,
-                    "attempt to add already present object {}",
-                    uri
+                    "attempt to add already present object {uri}"
                 )
             }
             DeltaError::ObjectHashMismatch { ref uri } => {
                 write!(
                     f,
-                    "local object {} has different hash",
-                    uri
+                    "local object {uri} has different hash"
                 )
             }
             DeltaError::ObjectRepeated { ref uri } => {
-                write!(f, "object appears multiple times: {}", uri)
+                write!(f, "object appears multiple times: {uri}")
             }
             DeltaError::LargeObject(ref uri) => {
-                write!(f, "object exceeds size limit: {}", uri)
+                write!(f, "object exceeds size limit: {uri}")
             }
             DeltaError::DeltaHashMismatch => {
                 write!(f, "delta file hash value mismatch")
             }
             DeltaError::Archive(ref err) => {
-                write!(f, "archive error: {}", err)
+                write!(f, "archive error: {err}")
             }
         }
     }
