@@ -50,7 +50,10 @@ const DEFAULT_EXPIRE: u64 = 7200;
 const DEFAULT_HISTORY_SIZE: usize = 10;
 
 /// The default for the RRDP timeout.
-const DEFAULT_RRDP_TIMEOUT: Duration = Duration::from_secs(300);
+const DEFAULT_RRDP_TIMEOUT: Duration = Duration::from_secs(600);
+
+/// The default for the RRDP read timeout.
+const DEFAULT_RRDP_READ_TIMEOUT: Duration = Duration::from_secs(10);
 
 /// The default for the RRDP TCP keepalive
 const DEFAULT_RRDP_TCP_KEEPALIVE: Duration = Duration::from_secs(60);
@@ -231,10 +234,22 @@ pub struct Config {
 
     /// RRDP timeout in seconds.
     ///
+    /// This is the timeout for completing the full HTTP request.
+    ///
     /// If this is None, no timeout is set.
     pub rrdp_timeout: Option<Duration>,
 
+    /// RRDP read timeout in seconds.
+    ///
+    /// This is the timeout for connect, read, and write calls within an
+    /// HTTP request.
+    ///
+    /// If this is None, no timeout is set.
+    pub rrdp_read_timeout: Option<Duration>,
+
     /// Optional RRDP connect timeout.
+    ///
+    /// If this is not set, the `rrdp_read_timeout` applies.
     pub rrdp_connect_timeout: Option<Duration>,
 
     /// Optional TCP keepalive duration for RRDP connections.
@@ -254,9 +269,6 @@ pub struct Config {
     /// RRDP HTTP User Agent.
     pub rrdp_user_agent: String,
 
-    /// Should we keep RRDP responses and if so where?
-    pub rrdp_keep_responses: Option<PathBuf>,
-
     /// Optional size limit for objects.
     pub max_object_size: Option<u64>,
 
@@ -283,6 +295,9 @@ pub struct Config {
     /// The refresh interval for repository validation.
     pub refresh: Duration,
 
+    /// The minimum refresh interval for repository validation.
+    pub min_refresh: Option<Duration>,
+
     /// The RTR retry inverval to be announced to a client.
     pub retry: Duration,
 
@@ -333,6 +348,9 @@ pub struct Config {
     /// The target to log to.
     pub log_target: LogTarget,
 
+    /// Should we log repository issues?
+    pub log_repository_issues: bool,
+
     /// The optional PID file for server mode.
     pub pid_file: Option<PathBuf>,
 
@@ -559,6 +577,16 @@ impl Config {
             };
         }
 
+        // rrdp_read_timeout
+        if let Some(value) = args.rrdp_read_timeout {
+            self.rrdp_read_timeout = if value == 0 {
+                None
+            }
+            else {
+                Some(Duration::from_secs(value))
+            };
+        }
+
         // rrdp_connect_timeout
         if let Some(value) = args.rrdp_connect_timeout {
             self.rrdp_connect_timeout = Some(Duration::from_secs(value))
@@ -591,11 +619,6 @@ impl Config {
             self.rrdp_proxies = list
         }
 
-        // rrdp_keep_responses
-        if let Some(path) = args.rrdp_keep_responses {
-            self.rrdp_keep_responses = Some(path)
-        }
-
         // max_object_size
         if let Some(value) = args.max_object_size {
             if value == 0 {
@@ -645,6 +668,11 @@ impl Config {
             self.log_level = LevelFilter::Error
         }
 
+        // log_repository_issues
+        if args.log_repository_issues {
+            self.log_repository_issues = true
+        }
+
         Ok(())
     }
 
@@ -725,6 +753,11 @@ impl Config {
             self.refresh = Duration::from_secs(value)
         }
 
+        // min-refresh
+        if let Some(value) = args.min_refresh {
+            self.min_refresh = Some(Duration::from_secs(value))
+        }
+
         // retry
         if let Some(value) = args.retry {
             self.retry = Duration::from_secs(value)
@@ -939,6 +972,13 @@ impl Config {
                     None => Some(DEFAULT_RRDP_TIMEOUT)
                 }
             },
+            rrdp_read_timeout: {
+                match file.take_u64("rrdp-read-timeout")? {
+                    Some(0) => None,
+                    Some(value) => Some(Duration::from_secs(value)),
+                    None => Some(DEFAULT_RRDP_READ_TIMEOUT)
+                }
+            },
             rrdp_connect_timeout: {
                 file.take_u64("rrdp-connect-timeout")?.map(Duration::from_secs)
             },
@@ -958,7 +998,6 @@ impl Config {
                 file.take_string_array("rrdp-proxies")?.unwrap_or_default()
             },
             rrdp_user_agent: DEFAULT_RRDP_USER_AGENT.to_string(),
-            rrdp_keep_responses: file.take_path("rrdp-keep-responses")?,
             max_object_size: {
                 match file.take_u64("max-object-size")? {
                     Some(0) => None,
@@ -987,6 +1026,9 @@ impl Config {
                     file.take_u64("refresh")?.unwrap_or(DEFAULT_REFRESH)
                 )
             },
+            min_refresh: {
+                file.take_u64("min-refresh")?.map(Duration::from_secs)
+            },
             retry: {
                 Duration::from_secs(
                     file.take_u64("retry")?.unwrap_or(DEFAULT_RETRY)
@@ -1034,6 +1076,9 @@ impl Config {
                 file.take_from_str("log-level")?.unwrap_or(LevelFilter::Warn)
             },
             log_target,
+            log_repository_issues: {
+                file.take_bool("log-repository-issues")?.unwrap_or(false)
+            },
             pid_file: file.take_path("pid-file")?,
             working_dir: file.take_path("working-dir")?,
             chroot: file.take_path("chroot")?,
@@ -1172,13 +1217,13 @@ impl Config {
             rrdp_max_delta_count: DEFAULT_RRDP_MAX_DELTA_COUNT,
             rrdp_max_delta_list_len: DEFAULT_RRDP_MAX_DELTA_LIST_LEN,
             rrdp_timeout: Some(DEFAULT_RRDP_TIMEOUT), 
+            rrdp_read_timeout: Some(DEFAULT_RRDP_READ_TIMEOUT), 
             rrdp_connect_timeout: None,
             rrdp_tcp_keepalive: Some(DEFAULT_RRDP_TCP_KEEPALIVE),
             rrdp_local_addr: None,
             rrdp_root_certs: Vec::new(),
             rrdp_proxies: Vec::new(),
             rrdp_user_agent: DEFAULT_RRDP_USER_AGENT.to_string(),
-            rrdp_keep_responses: None,
             max_object_size: Some(DEFAULT_MAX_OBJECT_SIZE),
             max_ca_depth: DEFAULT_MAX_CA_DEPTH,
             enable_bgpsec: false,
@@ -1186,6 +1231,7 @@ impl Config {
             dirty_repository: DEFAULT_DIRTY_REPOSITORY,
             validation_threads: Config::default_validation_threads(),
             refresh: Duration::from_secs(DEFAULT_REFRESH),
+            min_refresh: None,
             retry: Duration::from_secs(DEFAULT_RETRY),
             expire: Duration::from_secs(DEFAULT_EXPIRE),
             history_size: DEFAULT_HISTORY_SIZE,
@@ -1202,6 +1248,7 @@ impl Config {
             http_tls_cert: None,
             log_level: LevelFilter::Warn,
             log_target: LogTarget::default(),
+            log_repository_issues: false,
             pid_file: None,
             working_dir: None,
             chroot: None,
@@ -1368,6 +1415,13 @@ impl Config {
                 Some(value) => value.as_secs(),
             }
         );
+        insert_int(
+            &mut res, "rrdp-read-timeout",
+            match self.rrdp_read_timeout {
+                None => 0,
+                Some(value) => value.as_secs(),
+            }
+        );
         if let Some(timeout) = self.rrdp_connect_timeout {
             insert_int(&mut res, "rrdp-connect-timeout", timeout.as_secs());
         }
@@ -1397,11 +1451,6 @@ impl Config {
                 }).collect()
             )
         );
-        if let Some(path) = self.rrdp_keep_responses.as_ref() {
-            insert(
-                &mut res,"rrdp-keep-responses", format!("{}", path.display())
-            );
-        }
         insert_int(
             &mut res, "max-object-size",
             self.max_object_size.unwrap_or(0),
@@ -1412,6 +1461,9 @@ impl Config {
         insert(&mut res, "dirty", self.dirty_repository);
         insert_int(&mut res, "validation-threads", self.validation_threads);
         insert_int(&mut res, "refresh", self.refresh.as_secs());
+        if let Some(min_refresh) = self.min_refresh.as_ref() {
+            insert_int(&mut res, "min-refresh", min_refresh.as_secs());
+        }
         insert_int(&mut res, "retry", self.retry.as_secs());
         insert_int(&mut res, "expire", self.expire.as_secs());
         insert_int(&mut res, "history-size", self.history_size);
@@ -1492,6 +1544,7 @@ impl Config {
                 insert(&mut  res, "log-file", file.display().to_string());
             }
         }
+        insert(&mut res, "log-repository-issues", self.log_repository_issues);
         if let Some(ref file) = self.pid_file {
             insert(&mut res, "pid-file", file.display().to_string());
         }
@@ -1660,7 +1713,7 @@ impl FromStr for FilterPolicy {
             "reject" => Ok(FilterPolicy::Reject),
             "warn" => Ok(FilterPolicy::Warn),
             "accept" => Ok(FilterPolicy::Accept),
-            _ => Err(format!("invalid policy '{}'", s))
+            _ => Err(format!("invalid policy '{s}'"))
         }
     }
 }
@@ -1707,7 +1760,7 @@ impl FromStr for FallbackPolicy {
             "never" => Ok(FallbackPolicy::Never),
             "stale" => Ok(FallbackPolicy::Stale),
             "new" => Ok(FallbackPolicy::New),
-            _ => Err(format!("invalid policy '{}'", s))
+            _ => Err(format!("invalid policy '{s}'"))
         }
     }
 }
@@ -1824,10 +1877,14 @@ struct GlobalArgs {
     #[arg(long, value_name = "SECONDS")]
     rrdp_fallback_time: Option<u64>,
 
-    /// Timeout of network operation for RRDP (0 for none)
+    /// Timeout of HTTP request for RRDP (0 for none)
     #[arg(long, value_name = "SECONDS")]
     rrdp_timeout: Option<u64>,
 
+    /// Timeout of network operation for RRDP (0 for none)
+    #[arg(long, value_name = "SECONDS")]
+    rrdp_read_timeout: Option<u64>,
+
     /// Timeout for connecting to an RRDP server
     #[arg(long, value_name = "SECONDS")]
     rrdp_connect_timeout: Option<u64>,
@@ -1848,10 +1905,6 @@ struct GlobalArgs {
     #[arg(long, value_name = "URI")]
     rrdp_proxy: Option<Vec<String>>,
 
-    /// Keep RRDP responses in the given directory
-    #[arg(long, value_name = "PATH")]
-    rrdp_keep_responses: Option<PathBuf>,
-
     /// Maximum size of downloaded objects (0 for no limit)
     #[arg(long, value_name = "BYTES")]
     max_object_size: Option<u64>,
@@ -1897,6 +1950,10 @@ struct GlobalArgs {
     /// Log to this file
     #[arg(long, value_name = "PATH")]
     logfile: Option<String>,
+
+    /// Log repository issues
+    #[arg(long)]
+    log_repository_issues: bool,
 }
 
 
@@ -1905,10 +1962,14 @@ struct GlobalArgs {
 /// The server-related command line arguments.
 #[derive(Clone, Debug, Parser)]
 struct ServerArgs {
-    /// Refresh interval in seconds [default 3600]
+    /// Refresh interval in seconds [default 600]
     #[arg(long, value_name = "SECONDS")]
     refresh: Option<u64>,
 
+    /// Refresh interval in seconds [default 600]
+    #[arg(long, value_name = "SECONDS")]
+    min_refresh: Option<u64>,
+
     /// RTR retry interval in seconds [default 600]
     #[arg(long, value_name = "SECONDS")]
     retry: Option<u64>,
@@ -2046,8 +2107,7 @@ impl ConfigFile {
                 Ok(dir) => dir,
                 Err(err) => {
                     error!(
-                        "Fatal: Can't determine current directory: {}.",
-                        err
+                        "Fatal: Can't determine current directory: {err}."
                     );
                     return Err(Failed);
                 }
@@ -2544,7 +2604,7 @@ impl ConfigFile {
                 else {
                     first = false
                 }
-                print!("{}", key);
+                print!("{key}");
             }
             error!(".");
             Err(Failed)
@@ -2576,6 +2636,10 @@ fn facility_to_string(facility: Facility) -> String {
         LOG_CRON => "cron",
         LOG_AUTHPRIV => "authpriv",
         LOG_FTP => "ftp",
+        LOG_NTP => "ntp",
+        LOG_AUDIT => "audit",
+        LOG_ALERT => "alert",
+        LOG_CLOCK_DAEMON => "clockdaemon",
         LOG_LOCAL0 => "local0",
         LOG_LOCAL1 => "local1",
         LOG_LOCAL2 => "local2",
@@ -2636,6 +2700,7 @@ mod test {
             Config::default_validation_threads(),
         );
         assert_eq!(config.refresh, Duration::from_secs(DEFAULT_REFRESH));
+        assert_eq!(config.min_refresh, None);
         assert_eq!(config.retry, Duration::from_secs(DEFAULT_RETRY));
         assert_eq!(config.expire, Duration::from_secs(DEFAULT_EXPIRE));
         assert_eq!(config.history_size, DEFAULT_HISTORY_SIZE);
