@@ -1,11 +1,10 @@
 //! Handling of endpoints related to the status.
 
-use std::cmp;
-use chrono::{DateTime, Duration, Utc};
+use chrono::{Duration, Utc};
 use clap::{crate_name, crate_version};
 use crate::metrics::{
     HttpServerMetrics, PayloadMetrics, PublicationMetrics,
-    RtrClientMetrics, SharedRtrServerMetrics, VrpMetrics,
+    RtrServerMetrics, VrpMetrics,
 };
 use crate::payload::SharedHistory;
 use crate::utils::fmt::WriteOrPanic;
@@ -17,19 +16,19 @@ use super::response::{ContentType, Response, ResponseBuilder};
 //------------ handle_get ----------------------------------------------------
 
 pub async fn handle_get_or_head(
-    req: &Request,
+    req: Request,
     history: &SharedHistory,
     http: &HttpServerMetrics,
-    rtr: &SharedRtrServerMetrics,
-) -> Option<Response> {
+    rtr: &RtrServerMetrics,
+) -> Result<Response, Request> {
     let head = req.is_head();
     match req.uri().path() {
-        "/status" => Some(handle_status(head, history, http, rtr).await),
+        "/status" => Ok(handle_status(head, history, http, rtr).await),
         "/api/v1/status" => {
-            Some(handle_api_status(head, history, http, rtr).await)
+            Ok(handle_api_status(head, history, http, rtr).await)
         },
-        "/version" => Some(handle_version(head)),
-        _ => None
+        "/version" => Ok(handle_version(head)),
+        _ => Err(req)
     }
 }
 
@@ -40,14 +39,14 @@ async fn handle_status(
     head: bool,
     history: &SharedHistory,
     server_metrics: &HttpServerMetrics,
-    rtr_metrics: &SharedRtrServerMetrics,
+    rtr_metrics: &RtrServerMetrics,
 ) -> Response {
     let (metrics, serial, start, done, duration, unsafe_vrps) = {
         let history = history.read();
         (
             match history.metrics() {
                 Some(metrics) => metrics,
-                None => return Response::initial_validation(),
+                None => return Response::initial_validation(false),
             },
             history.serial(),
             history.last_update_start(),
@@ -78,16 +77,16 @@ async fn handle_status(
     );
 
     // serial
-    writeln!(res, "serial: {}", serial);
+    writeln!(res, "serial: {serial}");
 
     // last-update-start-at and -ago
     writeln!(res, "last-update-start-at:  {}", now - start);
-    writeln!(res, "last-update-start-ago: {}", start);
+    writeln!(res, "last-update-start-ago: {start}");
 
     // last-update-done-at and -ago
     if let Some(done) = done {
         writeln!(res, "last-update-done-at:   {}", now - done);
-        writeln!(res, "last-update-done-ago:  {}", done);
+        writeln!(res, "last-update-done-ago:  {done}");
     }
     else {
         writeln!(res, "last-update-done-at:   -");
@@ -96,7 +95,7 @@ async fn handle_status(
 
     // last-update-duration
     if let Some(duration) = duration {
-        writeln!(res, "last-update-duration:  {}", duration);
+        writeln!(res, "last-update-duration:  {duration}");
     }
     else {
         writeln!(res, "last-update-duration:  -");
@@ -239,39 +238,86 @@ async fn handle_status(
             );
         }
         if let Some(serial) = metrics.serial {
-            write!(res, ", serial={}", serial)
+            write!(res, ", serial={serial}")
         }
         writeln!(res)
     }
 
-    let detailed_rtr = rtr_metrics.detailed();
-    let rtr_metrics = rtr_metrics.read().await;
+    // RRDP logs
+    if metrics.has_rrdp_logs() {
+        writeln!(res, "rrdp-logs:");
+        for item in &metrics.rrdp {
+            if let Some(log) = &item.log_book {
+                writeln!(res, "    {}:", item.notify_uri);
+                for message in log {
+                    writeln!(res,
+                        "        [{}] {}",
+                        message.level,
+                        message.content,
+                    );
+                }
+            }
+        }
+    }
+
+    // rsync logs
+    if metrics.has_rsync_logs() {
+        writeln!(res, "rsync-logs:");
+        for item in &metrics.rsync {
+            if let Some(log) = &item.log_book {
+                writeln!(res, "    {}:", item.module);
+                for message in log {
+                    writeln!(res,
+                        "        [{}] {}",
+                        message.level,
+                        message.content,
+                    );
+                }
+            }
+        }
+    }
+
+    // pub_point_logs
+    if !metrics.pub_point_logs.is_empty() {
+        writeln!(res, "pub-point-issues:");
+        for (uri, book) in &metrics.pub_point_logs {
+            writeln!(res, "    {}:", uri);
+            for message in book {
+                writeln!(
+                    res, "        [{}] {}", message.level, message.content
+                );
+            }
+        }
+    }
+
+    let rtr = rtr_metrics.global();
 
     // rtr
     writeln!(res,
         "rtr-connections: {} current",
-        rtr_metrics.current_connections(),
+        rtr.current_connections(),
     );
     writeln!(res,
         "rtr-data: {} bytes sent, {} bytes received",
-        rtr_metrics.bytes_written(),
-        rtr_metrics.bytes_read()
+        rtr.bytes_written(),
+        rtr.bytes_read()
     );
 
-    if detailed_rtr {
+    if let Some(clients) = rtr_metrics.clients() {
         // rtr-clients
         writeln!(res, "rtr-clients:");
-        rtr_metrics.fold_clients(
-            RtrClientStatus::default(), RtrClientStatus::fold
-        ).for_each(|(addr, data)| {
-            write!(res, "    {}: connections={}, ", addr, data.connections);
-            if let Some(serial) = data.serial {
-                write!(res, "serial={}, ", serial);
+        clients.iter().for_each(|(addr, data)| {
+            write!(res,
+                "    {}: connections={}, ",
+                addr, data.current_connections()
+            );
+            if let Some(serial) = data.serial() {
+                write!(res, "serial={serial}, ");
             }
             else {
                 write!(res, "serial=N/A, ");
             }
-            if let Some(update) = data.updated {
+            if let Some(update) = data.updated() {
                 let update = Utc::now() - update;
                 write!(
                     res,
@@ -282,7 +328,7 @@ async fn handle_status(
             else {
                 write!(res, "updated=N/A, ");
             }
-            if let Some(update) = data.last_reset {
+            if let Some(update) = data.last_reset() {
                 let update = Utc::now() - update;
                 write!(
                     res,
@@ -295,8 +341,8 @@ async fn handle_status(
             }
             writeln!(res,
                 "reset-queries={}, serial-queries={}, read={}, written={}",
-                data.reset_queries, data.serial_queries,
-                data.bytes_read, data.bytes_written,
+                data.reset_queries(), data.serial_queries(),
+                data.bytes_read(), data.bytes_written(),
             );
         });
     }
@@ -327,14 +373,14 @@ async fn handle_api_status(
     head: bool,
     history: &SharedHistory,
     server_metrics: &HttpServerMetrics,
-    rtr_metrics: &SharedRtrServerMetrics,
+    rtr_metrics: &RtrServerMetrics,
 ) -> Response {
     let (metrics, serial, start, done, duration) = {
         let history = history.read();
         (
             match history.metrics() {
                 Some(metrics) => metrics,
-                None => return Response::initial_validation()
+                None => return Response::initial_validation(true)
             },
             history.serial(),
             history.last_update_start(),
@@ -348,8 +394,6 @@ async fn handle_api_status(
     }
 
     let now = Utc::now();
-    let detailed_rtr = rtr_metrics.detailed();
-    let rtr_metrics = rtr_metrics.read().await;
 
     let res = JsonBuilder::build(|target| {
         target.member_str("version",
@@ -433,6 +477,20 @@ async fn handle_api_status(
                         }
                         Err(_) => target.member_raw("duration", "null")
                     }
+                    if let Some(book) = &metrics.log_book {
+                        target.member_array("issues", |target| {
+                            for message in book {
+                                target.array_object(|target| {
+                                    target.member_str(
+                                        "level", message.level
+                                    );
+                                    target.member_str(
+                                        "messages", &message.content
+                                    );
+                                })
+                            }
+                        })
+                    }
                 })
             }
         });
@@ -486,72 +544,96 @@ async fn handle_api_status(
                             target.member_raw("snapshot_reason", "null");
                         }
                     }
+                    if let Some(book) = &metrics.log_book {
+                        target.member_array("issues", |target| {
+                            for message in book {
+                                target.array_object(|target| {
+                                    target.member_str(
+                                        "level", message.level
+                                    );
+                                    target.member_str(
+                                        "messages", &message.content
+                                    );
+                                })
+                            }
+                        })
+                    }
                 })
             }
         });
 
+        target.member_object("pubPointIssues", |target| {
+            for (uri, book) in &metrics.pub_point_logs {
+                target.member_array(uri, |target| {
+                    for message in book {
+                        target.array_object(|target| {
+                            target.member_str("level", message.level);
+                            target.member_str("message", &message.content);
+                        });
+                    }
+                });
+            }
+        });
+
         target.member_object("rtr", |target| {
+            let rtr = rtr_metrics.global();
             target.member_raw(
                 "currentConnections",
-                rtr_metrics.current_connections()
+                rtr.current_connections()
             );
             target.member_raw(
-                "bytesRead", rtr_metrics.bytes_read()
+                "bytesRead", rtr.bytes_read()
             );
             target.member_raw(
-                "bytesWritten", rtr_metrics.bytes_written()
+                "bytesWritten", rtr.bytes_written()
             );
 
-            if detailed_rtr {
+            if let Some(clients) = rtr_metrics.clients() {
                 target.member_object("clients", |target| {
-                    rtr_metrics.fold_clients(
-                        RtrClientStatus::default(), RtrClientStatus::fold
-                   ).for_each(
-                        |(addr, data)| {
-                            target.member_object(addr, |target| {
-                                target.member_raw(
-                                    "connections", data.connections
+                   clients.iter().for_each(|(addr, data)| {
+                        target.member_object(addr, |target| {
+                            target.member_raw(
+                                "connections", data.current_connections()
+                            );
+                            if let Some(serial) = data.serial() {
+                                target.member_raw("serial", serial);
+                            }
+                            else {
+                                target.member_raw("serial", "null");
+                            }
+                            if let Some(update) = data.updated() {
+                                target.member_str(
+                                    "updated",
+                                    update.format("%+")
                                 );
-                                if let Some(serial) = data.serial {
-                                    target.member_raw("serial", serial);
-                                }
-                                else {
-                                    target.member_raw("serial", "null");
-                                }
-                                if let Some(update) = data.updated {
-                                    target.member_str(
-                                        "updated",
-                                        update.format("%+")
-                                    );
-                                }
-                                else {
-                                    target.member_raw("updated", "null");
-                                }
-                                if let Some(update) = data.last_reset {
-                                    target.member_str(
-                                        "lastReset",
-                                        update.format("%+")
-                                    );
-                                }
-                                else {
-                                    target.member_raw("lastReset", "null");
-                                }
-                                target.member_raw(
-                                    "resetQueries", data.reset_queries
+                            }
+                            else {
+                                target.member_raw("updated", "null");
+                            }
+                            if let Some(update) = data.last_reset() {
+                                target.member_str(
+                                    "lastReset",
+                                    update.format("%+")
                                 );
-                                target.member_raw(
-                                    "serialQueries", data.serial_queries
-                                );
-                                target.member_raw(
-                                    "read", data.bytes_read
-                                );
-                                target.member_raw(
-                                    "written", data.bytes_written
-                                );
-                            })
-                        }
-                    );
-                });
+                            }
+                            else {
+                                target.member_raw("lastReset", "null");
+                            }
+                            target.member_raw(
+                                "resetQueries", data.reset_queries()
+                            );
+                            target.member_raw(
+                                "serialQueries", data.serial_queries()
+                            );
+                            target.member_raw(
+                                "read", data.bytes_read()
+                            );
+                            target.member_raw(
+                                "written", data.bytes_written()
+                            );
+                        })
+                    }
+                )});
             }
         });
 
@@ -667,53 +749,3 @@ fn handle_version(head: bool) -> Response {
     }
 }
 
-
-//------------ RtrClientStatus -----------------------------------------------
-
-// Helper struct to keep RTR client data.
-#[derive(Clone, Default)]
-struct RtrClientStatus {
-    connections: usize,
-    serial: Option<u32>,
-    updated: Option<DateTime<Utc>>,
-    last_reset: Option<DateTime<Utc>>,
-    reset_queries: u32,
-    serial_queries: u32,
-    bytes_read: u64,
-    bytes_written: u64,
-}
-
-impl RtrClientStatus {
-    fn fold(&mut self, client: &RtrClientMetrics) {
-        if client.is_open() {
-            self.connections += 1;
-        }
-        self.serial = match (
-            self.serial, client.serial().map(u32::from)
-        ) {
-            (Some(left), Some(right)) => Some(cmp::max(left, right)),
-            (Some(left), None) => Some(left),
-            (None, Some(right)) => Some(right),
-            (None, None) => None
-        };
-        self.updated = match (self.updated, client.updated()) {
-            (Some(left), Some(right)) => Some(cmp::max(left, right)),
-            (Some(left), None) => Some(left),
-            (None, Some(right)) => Some(right),
-            (None, None) => None
-        };
-        self.last_reset = match
-            (self.last_reset, client.last_reset())
-        {
-            (Some(left), Some(right)) => Some(cmp::max(left, right)),
-            (Some(left), None) => Some(left),
-            (None, Some(right)) => Some(right),
-            (None, None) => None
-        };
-        self.reset_queries += client.reset_queries();
-        self.serial_queries += client.serial_queries();
-        self.bytes_read += client.bytes_read();
-        self.bytes_written += client.bytes_written();
-    }
-}
-
