@@ -1,4 +1,4 @@
-/// The Routinator binary.
+//! The Routinator binary.
 
 use std::env::current_dir;
 use std::process::exit;
@@ -15,8 +15,7 @@ fn _main() -> Result<(), ExitError> {
         Ok(dir) => dir,
         Err(err) => {
             error!(
-                "Fatal: cannot get current directory ({}). Aborting.",
-                err
+                "Fatal: cannot get current directory ({err}). Aborting."
             );
             return Err(ExitError::Generic);
         }
