@@ -19,7 +19,7 @@ use std::str::FromStr;
 use std::sync::mpsc;
 use std::sync::Arc;
 use std::sync::mpsc::RecvTimeoutError;
-use std::time::{Duration, Instant};
+use std::time::{Duration, Instant, SystemTime};
 #[cfg(feature = "rta")] use bytes::Bytes;
 use clap::{Arg, Args, ArgAction, ArgMatches, FromArgMatches, Parser};
 use log::{error, info, warn};
@@ -33,7 +33,7 @@ use crate::{output, validity};
 use crate::config::Config;
 use crate::error::{ExitError, Failed, RunFailed};
 use crate::http::http_listener;
-use crate::metrics::{SharedRtrServerMetrics};
+use crate::metrics::RtrServerMetrics;
 use crate::output::{Output, OutputFormat};
 use crate::payload::{PayloadSnapshot, SharedHistory, ValidationReport};
 use crate::process::Process;
@@ -231,9 +231,9 @@ impl Server {
         warn!("Using config file {}.", process.config().config_file.display());
         process.setup_service(self.detach)?;
         let log = log.map(Arc::new);
-        let rtr_metrics = SharedRtrServerMetrics::new(
+        let rtr_metrics = Arc::new(RtrServerMetrics::new(
             process.config().rtr_client_metrics
-        );
+        ));
 
         let history = SharedHistory::from_config(process.config());
         let mut notify = NotifySender::new();
@@ -248,6 +248,14 @@ impl Server {
 
         process.drop_privileges()?;
 
+        // Load exceptions and quit if they are broken.
+        let Ok(mut exceptions) = LocalExceptions::load(
+            process.config(), true
+        ) else {
+            error!("Exiting.");
+            return Err(ExitError::Generic)
+        };
+
         let mut validation = Engine::new(process.config(), true)?;
         let runtime = process.runtime()?;
         let mut rtr = runtime.spawn(rtr);
@@ -258,58 +266,74 @@ impl Server {
         validation.ignite()?;
 
         let join = thread::spawn(move || {
+            let mut initial = true;
             let mut can_retry = true;
             let err = loop {
                 if let Some(log) = log.as_ref() {
                     log.start();
                 }
 
-                let timeout = match LocalExceptions::load(
-                    process.config(), true
-                ) {
-                    Ok(exceptions) => {
-                        match Self::process_once(
-                            process.config(), &validation, &history,
-                            &mut notify, exceptions,
-                        ) {
-                            Ok(()) => {
-                                history.read().refresh_wait()
-                            }
-                            Err(err) => {
-                                if err.should_retry() {
-                                    if can_retry {
-                                        if validation.sanitize().is_err() {
-                                            break Err(Failed)
-                                        }
-                                        info!(
-                                            "Validation failed but \
-                                             can be retried."
-                                        );
-                                        can_retry = false;
-                                        Duration::from_secs(0)
-                                    }
-                                    else {
-                                        error!(
-                                            "Retried validation failed again."
-                                        );
-                                        break Err(Failed);
-                                    }
-                                }
-                                else {
-                                    break Err(Failed);
-                                }
-                            }
-                        }
-
+                match LocalExceptions::load(process.config(), true) {
+                    Ok(new_exceptions) => {
+                        exceptions = new_exceptions;
                     }
                     Err(_) => {
-                        error!(
-                            "Failed to load exceptions. \
-                            Trying again in 10 seconds."
-                        );
-                        Duration::from_secs(10)
+                        warn!("Using previously loaded local exceptions.");
+                    }
+                }
+
+                let initial_run = initial;
+                initial = false;
+                let timeout = match Self::process_once(
+                    process.config(), &validation, &history,
+                    &mut notify, &exceptions, initial_run,
+                ) {
+                    Ok(()) => {
+                        // Immediately start a new run after the
+                        // initial run.
+                        if initial_run {
+                            info!(
+                                "Initial run complete, now starting \
+                                 normal run."
+                            );
+                            Duration::from_secs(0)
+                        }
+                        else {
+                            history.read().refresh_wait()
+                        }
+                    }
+                    Err(err) => {
+                        if err.should_retry() {
+                            if initial_run {
+                                info!(
+                                    "Retrying full validation run."
+                                );
+                                Duration::from_secs(0)
+                            }
+                            else if can_retry {
+                                if validation.sanitize().is_err() {
+                                    break Err(Failed)
+                                }
+                                info!(
+                                    "Validation failed but \
+                                     can be retried."
+                                );
+                                can_retry = false;
+                                Duration::from_secs(0)
+                            }
+                            else {
+                                error!(
+                                    "Retried validation failed again."
+                                );
+                                break Err(Failed);
+                            }
+                        }
+                        else {
+                            break Err(Failed);
+                        }
                     }
                 };
+
                 if let Some(log) = log.as_ref() {
                     log.flush();
                 }
@@ -401,32 +425,47 @@ impl Server {
         engine: &Engine,
         history: &SharedHistory,
         notify: &mut NotifySender,
-        exceptions: LocalExceptions,
+        exceptions: &LocalExceptions,
+        initial: bool,
     ) -> Result<(), RunFailed> {
         info!("Starting a validation run.");
         history.mark_update_start();
-        let (report, metrics) = ValidationReport::process(engine, config)?;
+        let (report, metrics) = ValidationReport::process(
+            engine, config, initial
+        )?;
         let must_notify = history.update(
-            report, &exceptions, metrics,
+            report, exceptions, metrics,
         );
+        history.mark_update_done();
         if log::max_level() >= log::Level::Info {
-            info!("Validation completed.");
-            let (metrics, serial) = {
+            let (metrics, serial, duration) = {
                 let history = history.read();
-                (history.metrics(), history.serial())
+                (
+                    history.metrics(),
+                    history.serial(),
+                    history.last_update_duration()
+                )
             };
+            match duration {
+                Some(duration) => {
+                    info!(
+                        "Validation completed in {} seconds.",
+                        duration.as_secs()
+                    );
+                }
+                None => info!("Validation completed.")
+            }
             if let Some(metrics) = metrics {
                 output::Summary::log(&metrics)
             }
             info!(
-                "New serial is {}.", serial
+                "New serial is {serial}."
             );
         }
         if must_notify {
             info!("Sending out notifications.");
             notify.notify();
         }
-        history.mark_update_done();
         Ok(())
     }
 }
@@ -435,6 +474,7 @@ impl Server {
 //------------ Vrps ----------------------------------------------------------
 
 /// Produce a list of Validated ROA Payload.
+#[derive(Debug)]
 pub struct Vrps {
     /// The destination to output the list to.
     ///
@@ -451,6 +491,9 @@ pub struct Vrps {
     /// Don’t update the repository.
     noupdate: bool,
 
+    /// Only update the repository if the cache is older than this.
+    update_after: Option<Duration>,
+
     /// Return an error on incomplete update.
     complete: bool,
 }
@@ -502,9 +545,20 @@ struct VrpsArgs {
     no_aspas: bool,
 
     /// Don't update the local cache
-    #[arg(short, long)]
+    #[arg(
+        short, long,
+        conflicts_with = "update_after",
+    )]
     noupdate: bool,
 
+    /// Only update the cache if the last run was at least this long ago.
+    #[arg(
+        short, long,
+        value_name = "MINTUES",
+        conflicts_with = "noupdate",
+    )]
+    update_after: Option<u32>,
+
     /// Return an error status on incomplete update
     #[arg(long)]
     complete: bool,
@@ -576,6 +630,11 @@ impl Vrps {
             format,
             output,
             noupdate: args.noupdate,
+            update_after: args.update_after.map(|minutes| {
+                // Safety: a u32 converted to a u64 multiplied by 60 always
+                //         fits.
+                Duration::from_secs(u64::from(minutes) * 60)
+            }),
             complete: args.complete,
         })
     }
@@ -590,6 +649,21 @@ impl Vrps {
     fn run(mut self, process: Process) -> Result<(), ExitError> {
         self.output.update_from_config(process.config());
         let mut engine = Engine::new(process.config(), !self.noupdate)?;
+
+        // Disable collector if update_after demands it. If anything goes
+        // wrong here, we simply keep the collector in place.
+        if let Some(duration) = self.update_after {
+            if let Some(status) = engine.store_status()? {
+                if let Ok(age) = SystemTime::from(
+                    status.last_update
+                ).elapsed() {
+                    if age < duration {
+                        engine.disable_collector()
+                    }
+                }
+            }
+        }
+
         engine.ignite()?;
         process.switch_logging(false, false)?;
         warn!("Using config file {}.", process.config().config_file.display());
@@ -599,7 +673,9 @@ impl Vrps {
             let mut once = false;
 
             loop {
-                match ValidationReport::process(&engine, process.config()) {
+                match ValidationReport::process(
+                    &engine, process.config(), false
+                ) {
                     Ok(res) => break res,
                     Err(err) => {
                         if err.should_retry() {
@@ -648,8 +724,7 @@ impl Vrps {
                 err.kind() != io::ErrorKind::BrokenPipe
             {
                 error!(
-                    "Failed to output result: {}",
-                    err
+                    "Failed to output result: {err}"
                 );
             }
             Err(ExitError::Generic)
@@ -806,8 +881,8 @@ impl Validate {
                 Ok(validity::RequestList::single(prefix, asn))
             }
             ValidateWhat::File(ref path) => {
-                let mut file = match fs::File::open(path) {
-                    Ok(file) => file,
+                let mut stream = match fs::File::open(path) {
+                    Ok(file) => io::BufReader::new(file),
                     Err(err) => {
                         error!(
                             "Failed to open input file '{}': {}'",
@@ -818,7 +893,7 @@ impl Validate {
                 };
                 if self.json {
                     validity::RequestList::from_json_reader(
-                        &mut file
+                        &mut stream
                     ).map_err(|err| {
                         error!(
                             "Failed to read input file '{}': {}'",
@@ -829,7 +904,7 @@ impl Validate {
                 }
                 else {
                     validity::RequestList::from_plain_reader(
-                        io::BufReader::new(file)
+                        &mut stream
                     ).map_err(|err| {
                         error!(
                             "Failed to read input file '{}': {}'",
@@ -846,7 +921,7 @@ impl Validate {
                     validity::RequestList::from_json_reader(
                         &mut file
                     ).map_err(|err| {
-                        error!("Failed to read input: {}'", err);
+                        error!("Failed to read input: {err}'");
                         ExitError::Generic
                     })
                 }
@@ -854,7 +929,7 @@ impl Validate {
                     validity::RequestList::from_plain_reader(
                         file
                     ).map_err(|err| {
-                        error!("Failed to read input: {}'", err);
+                        error!("Failed to read input: {err}'");
                         ExitError::Generic
                     })
                 }
@@ -869,7 +944,7 @@ impl Validate {
         engine.ignite()?;
         process.switch_logging(false, false)?;
         let (report, mut metrics) = ValidationReport::process(
-            &engine, process.config(),
+            &engine, process.config(), false,
         )?;
         let snapshot = report.into_snapshot(
             &LocalExceptions::load(process.config(), false)?,
@@ -892,8 +967,8 @@ impl Validate {
         let result = requests.validity(&snapshot);
         match self.output.as_ref() {
             Some(path) => {
-                let mut file = match fs::File::create(path) {
-                    Ok(file) => file,
+                let mut stream = match fs::File::create(path) {
+                    Ok(file) => io::BufWriter::new(file),
                     Err(err) => {
                         error!(
                             "Failed to open output file '{}': {}",
@@ -903,11 +978,12 @@ impl Validate {
                     }
                 };
                 let res = if self.json {
-                    result.write_json(&mut file)
+                    result.write_json(&mut stream)
                 }
                 else {
-                    result.write_plain(&mut file)
+                    result.write_plain(&mut stream)
                 };
+                let res = res.and_then(|_| stream.flush());
                 res.map_err(|err| {
                     error!(
                         "Failed to write to output file '{}': {}",
@@ -926,7 +1002,7 @@ impl Validate {
                     result.write_plain(&mut stdout)
                 };
                 res.map_err(|err| {
-                    error!("Failed to write output: {}", err);
+                    error!("Failed to write output: {err}");
                     ExitError::Generic
                 })
             }
@@ -1047,7 +1123,7 @@ impl ValidateDocument {
         match rta_validation.finalize() {
             Ok(rta) => {
                 for block in rta.as_resources().iter() {
-                    println!("{}", block);
+                    println!("{block}");
                 }
                 for block in rta.v4_resources().iter() {
                     println!("{}", block.display_v4());
@@ -1108,7 +1184,7 @@ impl Update {
         engine.ignite()?;
         process.switch_logging(false, false)?;
         let (_, metrics) = ValidationReport::process(
-            &engine, process.config(),
+            &engine, process.config(), false
         )?;
         if self.complete && !metrics.rsync_complete() {
             Err(ExitError::IncompleteUpdate)
@@ -1315,7 +1391,7 @@ impl Man {
                     }
                 };
                 if let Err(err) = file.write_all(MAN_PAGE) {
-                    error!("Failed to write to output file: {}", err);
+                    error!("Failed to write to output file: {err}");
                     return Err(Failed.into())
                 }
                 info!(
@@ -1327,7 +1403,7 @@ impl Man {
                 let out = io::stdout();
                 let mut out = out.lock();
                 if let Err(err) = out.write_all(MAN_PAGE) {
-                    error!("Failed to write man page: {}", err);
+                    error!("Failed to write man page: {err}");
                     return Err(Failed.into())
                 }
             }
@@ -1343,21 +1419,19 @@ impl Man {
         let mut file = NamedTempFile::new().map_err(|err| {
             error!(
                 "Can't display man page: \
-                 Failed to create temporary file: {}.",
-                err
+                 Failed to create temporary file: {err}."
             );
             Failed
         })?;
         file.write_all(MAN_PAGE).map_err(|err| {
             error!(
                 "Can't display man page: \
-                Failed to write to temporary file: {}.",
-                err
+                Failed to write to temporary file: {err}."
             );
             Failed
         })?;
         Command::new("man").arg(file.path()).status().map_err(|err| {
-            error!("Failed to run man: {}", err);
+            error!("Failed to run man: {err}");
             Failed
         }).and_then(|exit| {
             if exit.success() {
@@ -1395,14 +1469,14 @@ impl SignalListener {
             usr1: match signal(SignalKind::user_defined1()) {
                 Ok(usr1) => usr1,
                 Err(err) => {
-                    error!("Attaching to signal USR1 failed: {}", err);
+                    error!("Attaching to signal USR1 failed: {err}");
                     return Err(Failed)
                 }
             },
             usr2: match signal(SignalKind::user_defined2()) {
                 Ok(usr2) => usr2,
                 Err(err) => {
-                    error!("Attaching to signal USR2 failed: {}", err);
+                    error!("Attaching to signal USR2 failed: {err}");
                     return Err(Failed)
                 }
             },
